#!/usr/bin/perl -w

use strict;
use IO::Socket::INET;
use Data::Dumper;
use URI::Escape;

use constant HTTP_REQUEST_TIMEOUT => 10;
use constant DEBUG                => 0;
use constant SPEED_FUZZ           => 0.5;
$| = 1;

sub uri_escape_lower
{
	my $in = shift;
	$in = uri_escape($in);
	$in =~ s/%(\S\S)/"%".lc($1)/eg;    # lowercase
	return $in;
}

sub http_encode_hash
{
	my ($in) = @_;
	$in = unpack( "A20", pack( "H40", $in ) );
	$in = uri_escape_lower($in);
	return $in;
}

sub _decode
{
	my ($ref) = @_;

	my $cx = shift(@$ref);
	if ( $cx eq 'd' )
	{

		# -> A Dict : Cycle-Decode chunks until we get to an 'e'
		my $stolen_item = shift( @{$ref} );
		my %ret_hash    = ();
		while ( defined($stolen_item) && $stolen_item ne 'e' )
		{
			unshift( @{$ref}, $stolen_item );
			my $key = _decode($ref);
			my $val = _decode($ref);
			$ret_hash{$key} = $val;
			print "DICT + $key -> $val\n" if DEBUG;
			$stolen_item = shift( @{$ref} );    #NextOne
		}
		return \%ret_hash;
	}
	elsif ( $cx =~ /^\d$/ )
	{

		# -> LengthPrefixed string
		print "STRING\n" if DEBUG;
		my $string = undef;
		my $length = $cx;

		my $stolen_item = shift( @{$ref} );
		while ( defined($stolen_item) && $stolen_item ne ':' )
		{
			$length .= $stolen_item;
			$stolen_item = shift( @{$ref} );
		}
		for ( 1 .. $length )
		{
			$string .= shift( @{$ref} );
		}
		print "STRING + $string\n" if DEBUG;
		return $string;
	}
	elsif ( $cx =~ 'i' )
	{

		# -> An int, cycle until we find a non-int (e)
		my $fullnum = undef;
		while ( defined( my $append = shift( @{$ref} ) ) )
		{
			last if $append !~ /^\d$/;

			$fullnum .= $append;
		}
		print "NUM: $fullnum\n" if DEBUG;
		return $fullnum;
	}
	elsif ( $cx eq 'l' )
	{

		# -> A List (aka Array) ; Cycle-Decode until we got an 'e'
		my $stolen_item = shift( @{$ref} );
		my @ret_array   = ();
		while ( defined($stolen_item) && $stolen_item ne 'e' )
		{
			unshift( @{$ref}, $stolen_item );
			push( @ret_array, _decode($ref) );
			$stolen_item = shift( @{$ref} );
		}
		return \@ret_array;
	}
	elsif ( defined($cx) )
	{
		warn
"$0::Bitflu::Bencoding : Unexpected entry '$cx' found; Aborting parsing due to syntax error.\n";
		return undef;
	}
}

sub decode
{
	my ($string) = @_;
	my @chars = split( //, $string );
	return _decode( \@chars );
}

sub doHTTP
{
	my ($h) = @_;

	foreach (qw(stats_left stats_uploaded stats_downloaded))
	{
		unless ( defined $h->{$_} )
		{

			#stats_left has a good chance to be below zero if we
			#received even 1 bad chunk..
			warn "Fixing invalid value of $_";
			$h->{$_} = 0;
		}
	}

	my $peer_id      = uri_escape_lower( $h->{peer_id} );
	my $port         = int( $h->{port} );
	my $downloaded   = int( $h->{stats_downloaded} );
	my $uploaded     = int( $h->{stats_uploaded} );
	my $left         = int( $h->{stats_left} );
	my $URI          = $h->{uri};
	my $FULL_URI     = $h->{full_uri};
	my $torrent_hash = http_encode_hash( $h->{info_hash} );
	my $event        = $h->{event} || "";
	my $key          = $h->{key};
	my $ua           = $h->{ua};

	my ( $tracker_host, $tracker_port, $tracker_base ) = undef;
	my %ret = ();

	my $query = "GET ";

	if ($FULL_URI)
	{
		( $tracker_host, $tracker_port, $tracker_base ) =
		  $FULL_URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		$query .= "/$tracker_base HTTP/1.0\r\n";
		$query .= "Host: $tracker_host:$tracker_port\r\n";
		$query .= "\r\n";
	}
	else
	{

		#1: Decode tracker infos
		( $tracker_host, $tracker_port, $tracker_base ) =
		  $URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;

		#2: Create the GET request
		$query .=
		    "/$tracker_base"
		  . ( index( $tracker_base, "?" ) != -1 ? "&" : "?" )
		  . "info_hash=$torrent_hash";
		$query .= "&peer_id=$peer_id&port=$port";
		$query .= "&uploaded=$uploaded&downloaded=$downloaded&left=$left";
		$query .= "&event=$event" if $event;
		$query .= "&key=$key";
		$query .= "&compact=1&numwant=200&supportcrypto=1&no_peer_id=1";
		$query .= " HTTP/1.0\r\n";
		$query .= "User-Agent: $ua\r\n" if $ua;
		$query .= "Host: $tracker_host"
		  . ( $tracker_port != 80 ? ":$tracker_port" : "" ) . "\r\n";
		$query .= "\r\n";
	}

	return ( { ERROR => "Unable to extract parts" } ) unless $tracker_host;

	my $socket = IO::Socket::INET->new(
										PeerPort => $tracker_port,
										PeerHost => $tracker_host,
										Proto    => 'tcp',
										Timeout  => 2
	) or return ( { ERROR => "$!" } );

	$SIG{ALRM} = sub { die "TIMEOUT\n"; };
	alarm(HTTP_REQUEST_TIMEOUT);

	warn "<- h = " . ::Dumper($h)                if DEBUG;
	warn "<- host = $tracker_host:$tracker_port" if DEBUG;
	warn "<- $query"                             if DEBUG;

	eval {
		print $socket $query;

		my $data = <$socket>;
		print "-> $data" if DEBUG;
		( $ret{CODE} ) = $data =~ /^HTTP\/\d\.\d (\d{3})/;

		while (<$socket>)
		{
			print "-> $_" if DEBUG;
			s/\s+$//;
			if (/^([^:]+): (.+)$/)
			{
				$ret{HEADER}->{ lc($1) } = $2;
			}
			last if /^\s*$/;
		}

		$ret{BODY} = "";
		if ( $ret{HEADER}->{'content-length'} )
		{
			my $s = $ret{HEADER}->{'content-length'};
			while ($s)
			{
				my $data;
				$s -= $socket->read( $data, $s );
				$ret{BODY} .= $data;
			}
		}
		else
		{
			while (<$socket>)
			{
				$ret{BODY} .= $_;
			}
		}
	};

	delete $SIG{ALRM};
	alarm(0);

	if ($@)
	{
		return ( { ERROR => "Timed out" } );
	}
	else
	{
		return \%ret;
	}
}

################################################################################
#################################   My code    #################################
################################################################################

sub time_print
{
	my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
	  localtime(time);
	if ( length($min) < 2 ) { $min = "0" . $min }
	++$mon;
	if ( length($mon) < 2 ) { $mon = "0" . $mon }
	$year += 1900;
	print "[$mday-$mon-$year $hour:$min:$sec] ", @_;
}

sub pretty_size
{
	my $size  = shift;
	my $value = "";
	if ( $size > 1024 )
	{
		$value = "k";
		$size /= 1024;
	}
	if ( $size > 1024 )
	{
		$value = "M";
		$size /= 1024;
	}
	if ( $size > 1024 )
	{
		$value = "G";
		$size /= 1024;
	}
	if ( $size > 1024 )
	{
		$value = "T";
		$size /= 1024;
	}
	return int( $size * 100 ) / 100 . $value;
}

################################################################################
##############################   Initializing    ###############################
################################################################################

my $usage =
"USAGE $0 <announce_uri> <port> <info_hash> <peer_id[/UserAgent_string][/key_value]> <speed(bytes/sec)> <size> [max_download_size|\"completed\"]";

my $h;
$h->{uri}       = shift or die "No announce_uri param\n$usage";
$h->{port}      = shift or die "No port param\n$usage";
$h->{info_hash} = shift or die "No info_hash param\n$usage";
$h->{peer_id}   = shift or die "No peer_id param\n$usage";
( $h->{peer_id}, $h->{ua}, $h->{key} ) = split( /\//, $h->{peer_id} );
$h->{peer_id} = unpack( "A20", pack( "H40", $h->{peer_id} ) )
  if length( $h->{peer_id} ) == 40;
length( $h->{peer_id} ) == 20 or die "len(peer_id) != 20\n$usage";
$h->{key} = unpack( "H8", pack( "L", rand(0xFFFFFFFF) ) ) unless $h->{key};
my $speed = shift or die "No speed param\n$usage";
my $size  = shift or die "No size param\n$usage";
my $max_size = shift || $size;
if ($max_size ne "completed" && $max_size <= $size)
{
	die "max_download_size > size\n$usage";
}

my $work = 1;
my $result;
my $decoded_body;
my $sleep;
my $time     = 0;
my $max_time = 8 * 3600;    # no more 8 hours!

sub terminate
{
	$work = 0;
}
$SIG{INT}  = \&terminate;
$SIG{TERM} = \&terminate;

sub sighup
{
	time_print "HUP received\n";
}
$SIG{HUP} = \&sighup;

################################################################################
###############################   Main loop   ##################################
################################################################################

if ($max_size ne "completed")
{
	$h->{event}            = "started";    # started | completed | stopped
	$h->{stats_downloaded} = 0;
	$h->{stats_uploaded}   = 0;
	$h->{stats_left}       = $size;

	while ( $work && $time < $max_time )
	{
		$result = doHTTP($h);
		delete $h->{event};                # clear event - important!
		if ( $result->{ERROR} || $result->{CODE} != 200 )
		{
			# $result->{BODY} =~ tr/\x09\x0a\x0d\x20-\x7f/./c if $result->{BODY};
			die "result=" . Dumper($result);
		}
		$decoded_body = decode( $result->{BODY} );

		if ( $decoded_body->{peers} )
		{
			if ( ref $decoded_body->{peers} )
			{
				time_print "received ", scalar( @{ $decoded_body->{peers} } ),
				  " peers\n";
			}
			else
			{
				# compact form
				time_print "received ", length( $decoded_body->{peers} ) / 6,
				  " peers (compact)\n";
			}
		}

		if ( $decoded_body->{interval} )
		{
			time_print "sleep for ", $decoded_body->{interval}, " seconds\n";
			$sleep = sleep $decoded_body->{interval};
		}
		else
		{

			# important!
			# No interval field. May be the torrent removed from the tracker or we've got banned.
			time_print "No interval received! Terminating.\n";
			$work = 0;
			$sleep = 0;
		
			time_print "result=", Dumper($result);
			time_print "body=", Dumper($decoded_body);
		}

		$time += $sleep;

		$h->{stats_downloaded} +=
		  $sleep * $speed * ( 1 - SPEED_FUZZ + rand( SPEED_FUZZ * 2 ) );
		$h->{stats_downloaded} = $max_size if $h->{stats_downloaded} > $max_size;
		if ( $h->{stats_downloaded} >= $size )
		{
			$h->{stats_downloaded} = $size;
			$h->{event}            = "completed";
		}
		$h->{stats_uploaded} +=
		  $sleep * $speed * ( 1 - SPEED_FUZZ + rand( SPEED_FUZZ * 2 ) );
		$h->{stats_left} = $size - $h->{stats_downloaded};

		time_print pretty_size( $h->{stats_uploaded} ), "b uploaded, ",
		  pretty_size( $h->{stats_downloaded} ), "b downloaded, ",
		  pretty_size( $h->{stats_left} ),       "b left\n";
	}
}
else
{
	$h->{event}            = "completed";    # started | completed | stopped
	$h->{stats_downloaded} = 0;
	$h->{stats_uploaded}   = 0;
	$h->{stats_left}       = 0;

	while ( $work && $time < $max_time )
	{
		$result = doHTTP($h);
		delete $h->{event};                # clear event - important!
		if ( $result->{ERROR} || $result->{CODE} != 200 )
		{
			# $result->{BODY} =~ tr/\x09\x0a\x0d\x20-\x7f/./c if $result->{BODY};
			die "result=" . Dumper($result);
		}
		$decoded_body = decode( $result->{BODY} );

		if ( $decoded_body->{peers} )
		{
			if ( ref $decoded_body->{peers} )
			{
				time_print "received ", scalar( @{ $decoded_body->{peers} } ),
				  " peers\n";
			}
			else
			{
				# compact form
				time_print "received ", length( $decoded_body->{peers} ) / 6,
				  " peers (compact)\n";
			}
		}

		if ( $decoded_body->{interval} )
		{
			time_print "sleep for ", $decoded_body->{interval}, " seconds\n";
			$sleep = sleep $decoded_body->{interval};
		}
		else
		{

			# important!
			# No interval field. May be the torrent removed from the tracker or we've got banned.
			time_print "No interval received! Terminating.\n";
			$work = 0;
			$sleep = 0;
		
			time_print "result=", Dumper($result);
			time_print "body=", Dumper($decoded_body);
		}

		$time += $sleep;

		$h->{stats_uploaded} +=
		  $sleep * $speed * ( 1 - SPEED_FUZZ + rand( SPEED_FUZZ * 2 ) );

		time_print pretty_size( $h->{stats_uploaded} ), "b uploaded\n";
	}
}

################################################################################
############################   The last query   ################################
################################################################################

$h->{event} = "stopped";

$result = doHTTP($h);
if ( $result->{ERROR} || $result->{CODE} != 200 )
{
	# $result->{BODY} =~ tr/\x09\x0a\x0d\x20-\x7f/./c if $result->{BODY};
	die "result=" . Dumper($result);
}

time_print "Terminated\n";
